//
//  EAGLView.m
//  GetInBehind
//
//  Created by Stephen Blackheath on 9/06/09.
//  Copyright __MyCompanyName__ 2009. All rights reserved.
//



#import <QuartzCore/QuartzCore.h>
#import <OpenGLES/EAGLDrawable.h>

#import "EAGLView.h"

// Haskell callbacks
void (*fInit)(void);
void (*fDealloc)(void);


#define USE_DEPTH_BUFFER 0

EAGLView* thiz;

// A class extension to declare private methods
@interface EAGLView ()

@property (nonatomic, retain) EAGLContext *context;
@property (nonatomic, assign) NSTimer *animationTimer;

- (BOOL) createFramebuffer;
- (void) destroyFramebuffer;

@end


@implementation EAGLView

@synthesize context;
@synthesize animationTimer;
@synthesize animationInterval;


// You must implement this method
+ (Class)layerClass {
    return [CAEAGLLayer class];
}

- (GLuint)loadImageFile:(NSString *)name ofType:(NSString *)extension mipmap:(BOOL)mipmap
{
	UIImage *image;
	int width, height;
	CGImageRef cgImage;
	void *data;
	CGContextRef cgContext;
	CGColorSpaceRef colorSpace;
	GLenum err;
    GLuint texture = 0;
	
	image = [UIImage imageWithContentsOfFile:[[NSBundle mainBundle] pathForResource:name ofType:extension]];
	if (image == nil)
	{
		NSLog(@"Failed to load %@.%@", name, extension);
		return 0;
	}
	
	cgImage = [image CGImage];
	width = CGImageGetWidth(cgImage);
	height = CGImageGetHeight(cgImage);
	colorSpace = CGColorSpaceCreateDeviceRGB();

	data = malloc(width * height * 4);
	cgContext = CGBitmapContextCreate(data, width, height, 8, width * 4, colorSpace, kCGImageAlphaNoneSkipLast | kCGBitmapByteOrder32Big);
	if (cgContext != NULL)
	{
        int i;
        unsigned char* d = (unsigned char*)data;
		CGContextDrawImage(cgContext, CGRectMake(0.0f, 0.0f, width, height), cgImage);

          // Improve transparency
        for (i = 0; i < width * height * 4; i += 4) {
            if (d[i] == 0 && d[i+1] == 0 && d[i+2] == 0)
                d[i+3] = 0;
        }

		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
		
		if (mipmap)
			glGenerateMipmapOES(GL_TEXTURE_2D);
		
		err = glGetError();
		if (err != GL_NO_ERROR)
			NSLog(@"Error uploading texture. glError: 0x%04X", err);
		
		CGContextRelease(cgContext);
	}
	
	free(data);
	CGColorSpaceRelease(colorSpace);

    return texture;
}

//The GL view is stored in the nib file. When it's unarchived it's sent -initWithCoder:
- (id)initWithCoder:(NSCoder*)coder {
    thiz = self;
    if ((self = [super initWithCoder:coder])) {
        // Get the layer
        CAEAGLLayer *eaglLayer = (CAEAGLLayer *)self.layer;
        
        eaglLayer.opaque = YES;
        eaglLayer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:
                                        [NSNumber numberWithBool:NO],
                                        kEAGLDrawablePropertyRetainedBacking,
                                        kEAGLColorFormatRGBA8,
                                        kEAGLDrawablePropertyColorFormat, nil];

        context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context]) {
            [self release];
            return nil;
        }
        
        animationInterval = 1.0 / 60.0;
        
        _size = [eaglLayer bounds].size;
        _size.width = roundf(_size.width);
        _size.height = roundf(_size.height);
        fInit();
    }
    return self;
}

unsigned int newSprite(const char* name, unsigned int name_len)
{
    NSString* nsName = [[NSString alloc] initWithBytes:name length:name_len encoding:NSUTF8StringEncoding];
    // Create a texture from an image
    // First create a UIImage object from the data in a image file, and then extract the Core Graphics image
    unsigned int tex = [thiz loadImageFile:nsName ofType:@"png" mipmap:FALSE];
    [nsName release];
    return tex;
}

double getWidth(void)
{
    return thiz->_size.width;
}

double getHeight(void)
{
    return thiz->_size.height;
}

void drawRect(double x0, double y0, double x1, double y1)
{
    GLfloat verts[8];

    verts[0] = x0; verts[1] = y0;
    verts[2] = x1; verts[3] = y0;
    verts[4] = x0; verts[5] = y1;
    verts[6] = x1; verts[7] = y1;

    glVertexPointer(2, GL_FLOAT, 0, verts);
    glEnableClientState(GL_VERTEX_ARRAY);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

void drawSprite(unsigned int tex, double x, double y)
{
    const GLfloat squareVertices[] = {
        -22.0f, -22.0f,
        22.0f,  -22.0f,
        -22.0f,  22.0f,
        22.0f,   22.0f,
    };
    const GLfloat squareTexCoords[] = {
        0.0, 1.0,
        1.0, 1.0,
        0.0, 0.0,
        1.0, 0.0
    };

    glPushMatrix();
    glTranslatef(x, y, 0);

    glVertexPointer(2, GL_FLOAT, 0, squareVertices);
    glEnableClientState(GL_VERTEX_ARRAY);
    /*glColorPointer(4, GL_UNSIGNED_BYTE, 0, squareColors);
     glEnableClientState(GL_COLOR_ARRAY);*/
    
    glTexCoordPointer(2, GL_FLOAT, 0, squareTexCoords);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1.0);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glPopMatrix();
}

void (*fDraw)(void);

void setDraw(void (*f)())
{
    fDraw = f;
}

- (void)drawView {
    
    // Replace the implementation of this method to do your own custom drawing
    GLenum err;

    [EAGLContext setCurrentContext:context];
    
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //glOrthof(-1.0f, 1.0f, -1.5f, 1.5f, -1.0f, 1.0f);
    glOrthof(0, _size.width, 0, _size.height, -1, 1);
    /*glMatrixMode(GL_MODELVIEW);
    glRotatef(3.0f, 0.0f, 0.0f, 1.0f);*/
    glClearColor(0.0f, 0.25f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
        
    if (fDraw != NULL)
        fDraw();
    
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
    
	err = glGetError();
	if (err != GL_NO_ERROR)
		NSLog(@"Error in frame. glError: 0x%04X", err);
}


- (void)layoutSubviews {
    [EAGLContext setCurrentContext:context];
    [self destroyFramebuffer];
    [self createFramebuffer];
    [self drawView];
}


- (BOOL)createFramebuffer {
    
    glGenFramebuffersOES(1, &viewFramebuffer);
    glGenRenderbuffersOES(1, &viewRenderbuffer);
    
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(CAEAGLLayer*)self.layer];
    glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, viewRenderbuffer);
    
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);
    
    if (USE_DEPTH_BUFFER) {
        glGenRenderbuffersOES(1, &depthRenderbuffer);
        glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
        glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, backingWidth, backingHeight);
        glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);
    }
    
    if(glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES) {
        NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }
    
    return YES;
}


- (void)destroyFramebuffer {
    
    glDeleteFramebuffersOES(1, &viewFramebuffer);
    viewFramebuffer = 0;
    glDeleteRenderbuffersOES(1, &viewRenderbuffer);
    viewRenderbuffer = 0;
    
    if(depthRenderbuffer) {
        glDeleteRenderbuffersOES(1, &depthRenderbuffer);
        depthRenderbuffer = 0;
    }
}


- (void)startAnimation {
    self.animationTimer = [NSTimer scheduledTimerWithTimeInterval:animationInterval target:self selector:@selector(drawView) userInfo:nil repeats:YES];
}


- (void)stopAnimation {
    self.animationTimer = nil;
}


- (void)setAnimationTimer:(NSTimer *)newTimer {
    [animationTimer invalidate];
    animationTimer = newTimer;
}


- (void)setAnimationInterval:(NSTimeInterval)interval {
    
    animationInterval = interval;
    if (animationTimer) {
        [self stopAnimation];
        [self startAnimation];
    }
}


- (void)dealloc {
    
    [self stopAnimation];
    
    fDealloc();

    if ([EAGLContext currentContext] == context) {
        [EAGLContext setCurrentContext:nil];
    }
    
    glDeleteTextures(1, &sheepTexture);
    
    [context release];  
    [super dealloc];
}

- (CGPoint) convertPointFromViewToSurface:(CGPoint)point
{
	CGRect				bounds = [self bounds];
	
	return CGPointMake((point.x - bounds.origin.x) / bounds.size.width * _size.width, (point.y - bounds.origin.y) / bounds.size.height * _size.height);
}

void (*fTouchesBegan)(double x, double y);
void (*fTouchesMoved)(double x, double y);
void (*fTouchesEnded)(double x, double y);

void setTouchesBegan(void (*f)(double x, double y)) {fTouchesBegan = f;}
void setTouchesMoved(void (*f)(double x, double y)) {fTouchesMoved = f;}
void setTouchesEnded(void (*f)(double x, double y)) {fTouchesEnded = f;}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
	CGRect				bounds = [self bounds];
    UITouch*	touch = [[event touchesForView:self] anyObject];
    CGPoint pt;
	//Convert touch point from UIView referential to OpenGL one (upside-down flip)
	pt = [touch locationInView:self];
	pt.y = bounds.size.height - pt.y;
    pt = [self convertPointFromViewToSurface:pt];
    if (fTouchesBegan != NULL)
        fTouchesBegan(pt.x, pt.y);
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
	CGRect				bounds = [self bounds];
    UITouch*	touch = [[event touchesForView:self] anyObject];
    CGPoint pt;
	//Convert touch point from UIView referential to OpenGL one (upside-down flip)
	pt = [touch locationInView:self];
	pt.y = bounds.size.height - pt.y;
    pt = [self convertPointFromViewToSurface:pt];
    if (fTouchesMoved != NULL)
        fTouchesMoved(pt.x, pt.y);
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
	CGRect				bounds = [self bounds];
    UITouch*	touch = [[event touchesForView:self] anyObject];
    CGPoint pt;
	//Convert touch point from UIView referential to OpenGL one (upside-down flip)
	pt = [touch locationInView:self];
	pt.y = bounds.size.height - pt.y;
    pt = [self convertPointFromViewToSurface:pt];
    if (fTouchesEnded != NULL)
        fTouchesEnded(pt.x, pt.y);
}

@end
